// Copyright 2005 the Contributors, as shown in the revision logs.
// Licensed under the Apache Public Source License 2.0 ("the License").
// You may not use this file except in compliance with the License.

package edu.berkeley.sbp.tib;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.misc.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.chr.*;
import java.util.*;
import java.io.*;
import static edu.berkeley.sbp.misc.Demo.*;

public class TibDoc {
    /*
    public static Text lf()     { Chars ret = new Chars(); ret.text = "\n"; return ret; }
    public static Text cr()     { Chars ret = new Chars(); ret.text = "\r"; return ret; }
    public static Text emdash() { return new Entity("mdash"); }
    public static char urlescape(int a, int b) { return (char)(10*a+b); }


    // Template Classes //////////////////////////////////////////////////////////////////////////////

    public static abstract class Text implements ToHTML {
        public static final Class[] subclasses = new Class[] { Chars.class, URL.class, Email.class };
        public void toHTML(ToHTML.HTML sb) { }
        public static class TextString extends Text {
            public String text;
            public String tag() { return null; }
            public void toHTML(ToHTML.HTML sb) { sb.tag(tag(), text); }
        }
        public static class TextArray extends Text {
            public Text[] t;
            public String tag() { return null; }
            public void toHTML(ToHTML.HTML sb) { sb.tag(tag(), t); }
        }
    }


    // Structural //////////////////////////////////////////////////////////////////////////////

    public static class Doc implements ToHTML {
        public Header head;
        public Body   body;
        public void toHTML(ToHTML.HTML sb) { sb.tag("html", body); }
        public static class Header extends HashMap<String, Text[]> {
            public static class KeyVal { public String key; public Text[] val; }
            public void attrs(KeyVal[] KeyVals) { for(int i=0; i<KeyVals.length; i++) this.put(KeyVals[i].key, KeyVals[i].val); }
        }
        public static class Body implements ToHTML {
            public Section[] sections;
            public void toHTML(ToHTML.HTML sb) { sb.append(sections); }
            public static class Section implements ToHTML {
                public Text header;
                public Paragraph[] paragraphs;
                public void toHTML(ToHTML.HTML sb) {
                    sb.tag("h3", header);
                    sb.append(paragraphs);
                }
            }
        }
    }


    // Paragraph //////////////////////////////////////////////////////////////////////////////

    public static interface Paragraph extends ToHTML {
        public static final Class[] subclasses = new Class[] { Blockquote.class, P.class, HR.class };
        public static class HR                                implements Paragraph { public void toHTML(ToHTML.HTML sb) { sb.append("\n<hr>\n"); } }
        public static class P          extends Text.TextArray implements Paragraph { public String tag() { return "p"; } }
        public static class Blockquote extends Text.TextArray implements Paragraph { public String tag() { return "blockquote"; } }
    }

    public static abstract class List extends Text {
        public Text[][] points;
        public abstract String tag();
        public void toHTML(ToHTML.HTML sb) {
            sb.append("<"+tag()+">\n");
            for(Text[] t : points) sb.tag("li", t);
            sb.append("</"+tag()+">\n");
        }
    }
    public static class OL extends List { public String tag() { return "ol"; } }
    public static class UL extends List { public String tag() { return "ul"; } }



    // Tags //////////////////////////////////////////////////////////////////////////////

    public static class Chars         extends Text.TextString { }
    public static class Symbol        extends Text.TextString { }
    public static class Keyword       extends Text.TextString { public String tag() { return "tt"; } }
    public static class Subscript     extends Text.TextString { public String tag() { return "sub"; } }
    public static class Superscript   extends Text.TextString { public String tag() { return "super"; } }
    public static class Bold          extends Text.TextArray  { public String tag() { return "b"; } }
    public static class Smallcap      extends Text.TextArray  { public String tag() { return "sc"; } }
    public static class Strikethrough extends Text.TextArray  { public String tag() { return "strike"; } }
    public static class TT            extends Text.TextArray  { public String tag() { return "tt"; } }
    public static class Underline     extends Text.TextArray  { public String tag() { return "u"; } }
    public static class Italic        extends Text.TextArray  { public String tag() { return "i"; } }
    public static class Citation      extends Text.TextArray  { } // FIXME
    public static class Footnote      extends Text.TextArray  { } // FIXME
    public static class LineBreak     extends Text            { public void toHTML(ToHTML.HTML sb) { sb.tag("br"); } }
    public static class Today         extends Text            { }
    public static class Euro          extends Text            { public void toHTML(ToHTML.HTML sb) { sb.entity(8364); } }
    public static class Link          extends Text {
        public Text[] text;
        public URI href;
        public void toHTML(ToHTML.HTML sb) { sb.tag("a", new Object[] { "href", href }, text); }
    }
    public static class Entity        extends Text {
        public final String entity;
        public Entity(String entity) { this.entity = entity; }
        public void toHTML(ToHTML.HTML sb) { sb.entity(entity); }
    }


    // Network //////////////////////////////////////////////////////////////////////////////

    public static interface Host extends ToHTML {
        public static class DNS implements Host {
            public String[] part;
            public void toHTML(ToHTML.HTML sb) {
                for(int i=0; i<part.length; i++)
                    sb.append((i==0 ? "" : ".")+part[i]);
            }
        }
        public static class IP  implements Host {
            public int a, b, c, d;
            public void toHTML(ToHTML.HTML sb) { sb.append(a+"."+b+"."+c+"."+d); }
        }
    }

    public static interface URI extends ToHTML {
        public static final Class[] subclasses = new Class[] { URL.class, Email.class };
    }
    public static class URL extends Text    implements URI {
        public String method;
        public Login login;
        public Host host;
        public int port;
        public String path;
        public void toHTML(ToHTML.HTML sb) {
            sb.append(method);
            sb.append("://");
            // login.toHTML(sb);   FIXME
            host.toHTML(sb);
            // sb.append(":");     FIXME
            // sb.append(port);
            sb.append("/");
            sb.append(path);
        }
    }
    public static class Email extends Text  implements URI {
        public String user;
        public Host host;
        public void toHTML(ToHTML.HTML sb) {
            sb.append(user);
            sb.append('@');
            host.toHTML(sb);
        }
    }

    public static class Login {
        public String username;
        public String password;
        public void toHTML(ToHTML.HTML sb) {
            sb.append(username);
            sb.append(':');
            sb.append(password);
            sb.append('@');
        }        
    }




    */
    /*
        public static void prefix(PrintWriter p) {
            p.println("% generated by TIBDOC");
            for(int i=0; i<packages.length; i++) p.println("\\usemodule["+packages[i]+"]");
            p.println("\\setuppapersize[letter]");
            p.println("\\setuppagenumbering[location=]");
            p.println("\\setupcolors[state=start]");
            //"\\setupinteraction[title={Title},author={Me},"++
            //"subtitle={Deez Nutz},keywords={blargh},color=blue]\n" ++
            //"\\setuppublications[database={me},numbering=yes,sort=author]\n" ++
            p.println("\\setuphead[section][style={\\ss\\bfa},number=no,before=\\blank\\hairline\\nowhitespace]");
            p.println("\\definelayout[mypage][backspace=1.75in,cutspace=1.75in,width=5in]");
            p.println("\\setuplayout[mypage]");
            p.println("\\definetypeface[myface][rm][Xserif][Warnock Pro]");
            p.println("\\definetypeface[myface][tt][Xmono][CMU Typewriter Text Regular][default]");
            p.println("\\definetypeface[myface][ss][Xsans][Myriad Pro][default]");
            p.println("\\usesymbols[uni]");
            p.println("\\definesymbol[1][{\\USymbCharZapf{39}{164}}]");
            p.println("\\setupbodyfont[myface, 11pt]");
            p.println("\\setupwhitespace[7pt]");
            p.println("\\def\\MyDroppedCaps%");
            p.println("    {\\DroppedCaps");
            p.println("        {} {Serif} {2\\baselineskip} {2pt} {1\\baselineskip} {2}}");
            p.println("\\starttext");
            p.println("\\switchtobodyfont[16pt]\\midaligned{\\ss\\bfa{Title}}\\switchtobodyfont[10pt]");
            p.println("\\midaligned{Adam Megacz}\n\n\\nowhitespace\\midaligned{\\tt{adam@megacz.com}}");
            p.println("\\blank[1cm,force]");
            //p.println("\\defineparagraphs[mypar][n=2,before={\\blank},after={\\blank}");
            //p.println("\\setupparagraphs[mypar][1][width=.45\\textwidth");
            //p.println("\\setupparagraphs[mypar][2][width=.55\\textwidth");
            //p.println("\\startmypa");
            //p.println("\\switchtobodyfont[sma");
        }

        public static void suffix(PrintWriter p) {
            p.println("\\stoptext");
        }
        static String[] packages = new String[] { "supp-fun", "bib", "href" };
    }

    // ConTex

module Contex where
import Data.Array.IArray
import Data.Char
import Util
import Lexer
import IR
import Data.List
import Beautify

toContex ll = prefix ++ (concatMap tl ll) ++ suffix
 where
  packages                         = [ "[supp-fun]",
                                       "[bib]",
                                       "[href]" ]
  prefix                           = (concatMap (\x -> "\\usemodule"++x++"\n") packages) ++
                                     "\\setuppapersize[letter]\n" ++
                                     "\\setuppagenumbering[location=]\n" ++
                                     "\\setupcolors[state=start]\n" ++
                                     --"\\setupinteraction[title={Title},author={Me},"++
                                     --"subtitle={Deez Nutz},keywords={blargh},color=blue]\n" ++
                                     --"\\setuppublications[database={me},numbering=yes,sort=author]\n" ++
                                     "\\setuphead[section][style={\\ss\\bfa},\n" ++
                                     "                     number=no,\n" ++
                                     "                     before=\\blank\\hairline\\nowhitespace,\n" ++
                                     "                     ]\n" ++
                                     "\\definelayout[mypage][\n" ++
                                     " backspace=1.75in, % the space for margin notes\n" ++
                                     " cutspace=1.75in, % the space for right margin notes\n" ++
                                     " width=5in" ++
                                     "]\n" ++
                                     "\\setuplayout[mypage]\n" ++
                                     "\\definetypeface[myface][rm][Xserif][Warnock Pro]\n" ++
                                     "\\definetypeface[myface][tt][Xmono][CMU Typewriter Text Regular][default]\n" ++
                                     "\\definetypeface[myface][ss][Xsans][Myriad Pro][default]\n" ++
                                     "\\usesymbols[uni]\n" ++
                                     "\\definesymbol[1][{\\USymbCharZapf{39}{164}}]\n" ++
                                     "\\setupbodyfont[myface, 11pt]\n" ++
                                     "\\setupwhitespace[7pt]\n" ++
                                     "\\def\\MyDroppedCaps%\n" ++
                                     "    {\\DroppedCaps\n" ++
                                     "        {} {Serif} {2\\baselineskip} {2pt} {1\\baselineskip} {2}}\n" ++
                                     "\\starttext\n" ++
                                     "\\switchtobodyfont[16pt]\\midaligned{\\ss\\bfa{Hi5 Replicated Server Infrastructure}}\\switchtobodyfont[10pt]\n"++ 
                                     "\\midaligned{Adam Megacz}\n\n\\nowhitespace\\midaligned{\\tt{adam@megacz.com}}\n\n"++
                                     "\\blank[1cm,force]\n" ++
                                     "\\defineparagraphs[mypar][n=2,before={\\blank},after={\\blank}]\n"++
                                     "\\setupparagraphs[mypar][1][width=.45\\textwidth]\n"++
                                     "\\setupparagraphs[mypar][2][width=.55\\textwidth]\n"++
                                     "\\startmypar"++
                                     "\\switchtobodyfont[small]\n"
  suffix                           = "\n\\stoptext\n"
  addItem i                        = "\\item " ++ (striptrail $ boost 8 $ wrap $ striplead $ tl i)
  escapify []                      = []
  escapify ('%':t)                 = '\\':'%':(escapify t)
  escapify ('$':t)                 = '\\':'$':(escapify t)
  escapify (h:t)                   = h:(escapify t)
  escapeMath s                     = s
  tl (Special _ BulletList l)      = "\\startitemize[symbol]\n" ++ (concatMap addItem l) ++ "\\stopitemize\n"
  tl (Special _ NumberList l)      = "\\startitemize[symbol]\n" ++ (concatMap addItem l) ++ "\\stopitemize\n"
  tl (Special _ Section (h:t))     = "\\section{"++(tl h)++"}\n"++(concatMap tl t)
  tl (Special _ NumberedSection (h:t)) = "\\section{"++(tl h)++"}\n"++(concatMap tl t)
  tl (Special _ (Glyph EmDash) []) = "{\\emdash}"
--tl (Special _ Superscript l)     = 
--tl (Special _ Subscript l)       = 
--tl (Special _ (Image u) l)       = 
--tl (Special _ (Cite u) l)        = 
  tl (Special _ BlockQuote l)      = "\n\n\\startquote\n     "++
                                     (striptrail $ boost 4 $ wrap $ striplead $ concatMap tl l)++"\n\\stopquote\n\n"
  tl (Special _ HorizontalRule []) = "\\\\\\rule{4in}{0.5pt}\\\\"
  tl (Special _ Italic l)          = "{\\it{"++(concatMap tl l)++"}}"
  tl (Special _ Bold l)            = "{\\bf{"++(concatMap tl l)++"}}"
  tl (Special _ DropCap (h:t))     = "\\MyDroppedCaps{"++(tl h)++"}{\\sc "++(concatMap tl t)++"}"
  tl (Special _ Typewriter l)      = "{\\tt{"++(concatMap tl l)++"}}"
  tl (Special _ StrikeThrough l)   = "" --"\\sout{"++(concatMap tl l)++"}"
  tl (Special _ Quotes l)          = "``"++(concatMap tl l)++"''"
  tl (Special _ Underline l)       = "" --"\\uline{"++(concatMap tl l)++"}"
  tl (Special _ Underline2 l)      = "" --"\\uuline{"++(concatMap tl l)++"}"
  tl (Special _ (Math s) [])       = "\\placeformula\n$$\n" ++ (escapeMath s) ++ "\n$$\n"
  tl (Special _ Footnote l)        = "\\footnote{"++(concatMap tl l)++"}"
  tl (Special _ Float l)           = "" --"\n\n\\begin{wrapfigure}{r}{0.4\\textwidth} \\framebox[0.4\\textwidth]{"++
                                     --(concatMap tl l)++"} \\label{x}\\end{wrapfigure}\n\n"
--tl (Special _ Figure l)          = "\\placefigure[][fig:church]{}{"++(concatMap tl l)++"}"
  tl (Special _ Figure l)          = "\\startnarrower\n"++(concatMap tl l)++"\n\\stopnarrower\n"
  tl (Special _ (Link u) l)        = "\\href{"++(escapify u)++"}{"++(concatMap tl l)++"}"
  tl (Special _ (Verbatim s) [])   = "\\starttyping\n"++s++"\n\\stoptyping\n"
--  tl (Special _ TwoColumn l)       = "\\startcolumns[n=2]\n"++(concatMap tl l)++"\\stopcolumns"
--  tl (Special _ Title l)           = ""--"\\title{"++(concatMap tl l)++"}\n\\maketitle\n\n\n"
  tl (Special _ Abstract l) =
      "\\midaligned{\\ss\\bfa Abstract}\\par\n " ++
      "\n\n"++(concatMap tl l)++"\\mypar" ++
      "\\switchtobodyfont[8pt]{\\ss{\\placecontent}}\\switchtobodyfont[normal]\\stopmypar\n\n\\blank[1cm,force]"
  tl (Special _ (Command c) l)     = "\\"++c++"["++(concatMap tl l)++"]"
  tl (Special _ t l)               = error $ "formatting code "++(show t)++" not supported on {"++(concatMap show l)++"})"
  tl (WS _)                        = " "
  tl (BlankLine _)                 = "\n\n"
  tl (Block _ l)                   = concatMap tl l
  tl (Letter _ c)                  = escapify [c]
  tl z                             = (show z)




    */

    // Main //////////////////////////////////////////////////////////////////////////////

    public static class Dump implements Reflection.Show {
        public String toString() { return Reflection.show((Reflection.Show)this); }
    }

    public static class TD {

        public @bind.as static class Doc extends Dump {
            public @arg Header head;
            public @arg Body body;
        }

        public @bind.as static class Header extends Dump {
            public @arg KeyVal[] attrs;
            // FIXME: it would be nice to be able to
            // void KeyVal(String, String) { ... } imperatively
        }
        
        public @bind.as static class Body extends Dump {
            public Section[] sections;
        }
        
        public @bind.as("Section") static class Section extends Dump {
            public String      header;
            public Paragraph[] paragraphs;
        }
        
        public @bind.as static class KeyVal extends Dump {
            public @arg String key;
            public @arg Object val;
        }

        public abstract static class Paragraph extends Dump implements ToHTML { }

        public @bind.as("P") static class P extends Paragraph {
            public Text[] text;
            public P() { }
            public P(Text[] text) { this.text = text; }
            public void toHTML(HTML h) { if (text != null) for (Text t : text) if (t != null) t.toHTML(h); }
            public String toString() {
                StringBuffer sb = new StringBuffer();
                ToHTML.HTML h = new ToHTML.HTML(sb);
                toHTML(h);
                return sb.toString();
            }
        }

        public @bind.as("HR") static class HR extends Paragraph {
            public void toHTML(HTML h) { h.tag("hr"); }
        }

        public @bind.as("Blockquote") static class Blockquote extends Paragraph {
            Text[] text;
            public void toHTML(HTML h) { h.tag("blockquote", new P(text)); }
        }

        public abstract static class Text extends Dump implements ToHTML { }
        public @bind.as static class Chars extends Text {
            public String text;
            public Chars() { }
            public Chars(String text) { this.text = text; }
            public void toHTML(HTML h) { h.appendText(" " + text + " "); }
            public String toString() { return text; }
        }
        public @bind.as static class Block extends Text {
            public Text[] text;
            public void toHTML(HTML h) { for(Text t : text) t.toHTML(h); }
        }
        public static class TextWrap extends Text {
            public Text text;
            public void toHTML(HTML h) {
                if (htmlTag()!=null)
                    h.tag(htmlTag(), htmlTagParams(), text);
                else
                    text.toHTML(h);
            }
            public String   htmlTag() { return null; }
            public Object[] htmlTagParams() { return null; }
        }
        public static @bind.as class Verbatim   extends Text { public char[] c; public void toHTML(HTML h) { } }
        //public @bind.as class Blockquote extends TextWrap { }
        public static @bind.as class Underline extends TextWrap { public String htmlTag() { return "u"; } }
        public static @bind.as class Footnote extends TextWrap { public String htmlTag() { return "small"; } }
        public static @bind.as class TT extends TextWrap { public String htmlTag() { return "tt"; } }
        //public @bind.as class Citation extends Text {       "[" word "]" }
        public static @bind.as class Strikethrough extends TextWrap { public String htmlTag() { return "strikethrough"; } }
        public static @bind.as class Superscript extends TextWrap { public String htmlTag() { return "sup"; } }
        public static @bind.as class Subscript extends TextWrap { public String htmlTag() { return "sub"; } }
        public static @bind.as class Smallcap extends TextWrap { public String htmlTag() { return "sc"; } }
        public static @bind.as class Keyword extends TextWrap { public String htmlTag() { return "sc"; } }
        public static @bind.as class Bold extends TextWrap { public String htmlTag() { return "b"; } }
        public static @bind.as class Italic extends TextWrap { public String htmlTag() { return "i"; } }

        public abstract static class Command extends Text { }
        public static @bind.as class Today extends Command { public void toHTML(HTML h) { } }
        public static @bind.as class LineBreak extends Command { public void toHTML(HTML h) { h.tag("br"); } }

        public abstract static class Glyph extends Text { }
        public static @bind.as("emdash") class Emdash extends Glyph { public void toHTML(HTML h) { h.append("&emdash;"); } }

        public static class Link extends Text {
            public Text[] t;
            public Url u;
            public @bind.as("LinkText") Link(Text[] t, Url u)  { this.t = t; this.u = u; }
            public @bind.as("LinkChars") Link(String s, Url u) { this(new Text[] { new Chars(s) }, u); }
            public void toHTML(HTML h) {
                h.tag("a",
                      new Object[] { "href", u==null ? "" : u.toString() },
                      new P(t));
            }
        }

        public static class Host {
            public String name;
            public String toString() { return name; }
            public @bind.as("DNS") Host(String[][] parts) {
                name = "";
                for(String[] s : parts) {
                    for(String ss : s)
                        name += ss;
                    name += ".";
                }
            }
            public @bind.as("IP")  Host(int a, int b, int c, int d) { name = a+"."+b+"."+c+"."+d; }
        }

        public static class Url extends Text {
            public String   method;
            public Host     host;
            public String   user;
            public String   pass;
            public String   port;
            public String   path;
            public @bind.as("URL") Url(String method, String[] login, Host host, String port, String path) {
                this.method = method;
                this.user = login==null ? null : login.length >= 1 ? login[0] : null;
                this.pass = login==null ? null : login.length >= 2 ? login[1] : null;
                this.host = host;
                this.port = port;
                this.path = path;
            }
            public @bind.as("Mailto") Url(String email) { this("mailto", null, null, "25", email); }
            public void toHTML(HTML h) { new Link(toString(), this).toHTML(h); }
            public String toString() {
                return method + "://" + host + "/" + path;
            }
        }
        public static @bind.as("lf")        String lf() { return "\r"; }
        public static @bind.as("cr")        String cr() { return "\n"; }
        public static @bind.as("\"\"")      String empty() { return ""; }
        public static @bind.as("urlescape") String urlescape(char a, char b) { return ((char)((a-'0') * 16 + (b-'0')))+""; }
    }

    public static void main(String[] s) throws Exception {
        try {

            Demo.ReflectiveMeta m =
                new Demo.ReflectiveMeta(TibDoc.TD.class,
                                        new Class[] {
                                            TibDoc.TD.Doc.class,
                                            TibDoc.TD.Header.class,
                                            TibDoc.TD.Section.class,
                                            TibDoc.TD.Url.class,
                                            TibDoc.TD.Host.class,
                                            TibDoc.TD.Link.class,
                                            TibDoc.TD.Body.class,
                                            TibDoc.TD.Paragraph.class,
                                            TibDoc.TD.P.class,
                                            TibDoc.TD.HR.class,
                                            TibDoc.TD.Blockquote.class,
                                            TibDoc.TD.KeyVal.class,
                                            TibDoc.TD.Text.class,
                                            TibDoc.TD.TextWrap.class,
                                            TibDoc.TD.Verbatim.class,
                                            TibDoc.TD.Blockquote.class,
                                            TibDoc.TD.Underline.class,
                                            TibDoc.TD.Footnote.class,
                                            TibDoc.TD.TT.class,
                                            TibDoc.TD.Strikethrough.class,
                                            TibDoc.TD.Superscript.class,
                                            TibDoc.TD.Subscript.class,
                                            TibDoc.TD.Smallcap.class,
                                            TibDoc.TD.Bold.class,
                                            TibDoc.TD.Keyword.class,
                                            TibDoc.TD.Italic.class,
                                            TibDoc.TD.Today.class,
                                            TibDoc.TD.LineBreak.class,
                                            TibDoc.TD.Chars.class,
                                            TibDoc.TD.Emdash.class,
                                        });
            Tree<String> res = new CharParser(MetaGrammar.make()).parse(new FileInputStream(s[0])).expand1();
            MetaGrammar.Meta.MetaGrammarFile mgf = m.new MetaGrammarFile(res);
            MetaGrammar.BuildContext bc = new MetaGrammar.BuildContext(mgf);
            Union tibgram = mgf.get("s").build(bc);

            System.err.println("parsing " + s[1]);
            Tree t = new CharParser(tibgram).parse(new Tib(new FileInputStream(s[1]))).expand1();
            System.out.println("tree:\n" + t.toPrettyString());
            
            Object result = ((Functor)t.head()).invoke(t);
            System.out.println((TD.Doc)result);
            /*
            System.out.println("parsing " + s[0]);
            Tree<String> res = new CharParser(MetaGrammar.make()).parse(new FileInputStream(s[0])).expand1();
            MetaGrammar gram = new Tib.Grammar(TibDoc.class);
            gram = (MetaGrammar)gram.walk(res);
            System.out.println("\nparsing " + s[1]);
            Forest f = new CharParser(gram.done()).parse(new Tib(new FileInputStream(s[1])));
            System.out.println();
            System.out.println(f.expand1().toPrettyString());
            System.out.println();


            FileOutputStream fos = new FileOutputStream("out.html");
            PrintWriter p = new PrintWriter(new OutputStreamWriter(fos));
            p.println(sb);
            p.flush();
            p.close();
            */
        } catch (Ambiguous a) {
            FileOutputStream fos = new FileOutputStream("/Users/megacz/Desktop/out.dot");
            PrintWriter p = new PrintWriter(new OutputStreamWriter(fos));
            GraphViz gv = new GraphViz();
            a.ambiguity.toGraphViz(gv);
            gv.dump(p);
            p.flush();
            p.close();
            a.printStackTrace();
            
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

}

